﻿-- Random UV XForm by Element
-- Applies a randomized transform to each individual element in a mesh.  Inspired by Railclone's UVW XForm Modifier.
-- By Avery Brown for Brooklyn Digital Foundry in December 2019

-- License:

-- Copyright 2019 Avery Brown

-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
-- documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
-- the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
-- to permit persons to whom the Software is furnished to do so, subject to the following conditions:

-- The above copyright notice and this permission notice shall be included in all copies or 
-- substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
-- BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

plugin simpleMeshMod UVWByElement
	name:"UV By Element"
	classID:#(0x5b86d01b, 0x61c4fa1b)
	category:"BDF"
	version: 1
(
	parameters main rollout:params
	(
		offset_on ui:ckbx_offset_on default:false type:#boolean animatable:false
		u_min_offset ui:spn_u_min_offset default:-1.0 type:#float animatable:false
		u_max_offset ui:spn_u_max_offset default:1.0 type:#float animatable:false
		v_min_offset ui:spn_v_min_offset default:-1.0 type:#float animatable:false
		v_max_offset ui:spn_v_max_offset default:1.0 type:#float animatable:false
		offset_inc ui:spn_offset_increment default:0.0 type:#float animatable:false
		
		rotation_on ui:ckbx_rotation_on default:false type:#boolean animatable:false
		min_rotation ui:spn_min_rotation default:-360.0 type:#float animatable:false
		max_rotation ui:spn_max_rotation default:360.0 type:#float animatable:false
		rotation_inc ui:spn_rotation_increment default:0.0 type:#float animatable:false
		
		scale_on ui:ckbx_scale_on default:false type:#boolean animatable:false
		u_min_scale ui:spn_u_min_scale default:1.0 type:#float animatable:false
		u_max_scale ui:spn_u_max_scale default:1.0 type:#float animatable:false
		v_min_scale ui:spn_v_min_scale default:1.0 type:#float animatable:false
		v_max_scale ui:spn_v_max_scale default:1.0 type:#float animatable:false
		scale_inc ui:spn_scale_increment default:0.0 type:#float animatable:false
		scale_aspect_lock ui:ckbx_scale_aspect_lock default:true type:#boolean animatable:false
		
		ch ui:spn_channel default:1 type:#integer animatable:false
		rseed ui:spn_seed default:12345 type:#integer animatable:false
	) 

	rollout params "Parameters"
	(
		group "Translation" 
		(
			checkbox ckbx_offset_on "Enabled"
			spinner spn_u_min_offset "U: Min:" range:[-1000.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_u_max_offset "Max:" range:[-1000.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_v_min_offset "V: Min:" range:[-1000.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_v_max_offset "Max:" range:[-1000.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_offset_increment "Increment:" range:[0.0, 1000.0, 0.1]  type:#float
		)
		
		group "Rotation" (
			checkbox ckbx_rotation_on "Enabled"
			spinner spn_min_rotation "Min:" range:[-360.0, 360.0, 15] across: 2 type:#float
			spinner spn_max_rotation "Max:" range:[-360.0, 360.0, 15] across: 2 type:#float
			spinner spn_rotation_increment "Increment:" range:[0.0, 360.0, 5.0]  type:#float
		)
		
		group "Scale" (
			checkbox ckbx_scale_on "Enabled"
			spinner spn_u_min_scale "U: Min:" range:[0.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_u_max_scale "Max:" range:[0.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_v_min_scale "V: Min:" range:[0.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_v_max_scale "Max:" range:[00.0, 1000.0, 0.1] across: 2 type:#float
			spinner spn_scale_increment "Increment:" range:[0.0, 1000.0, 0.1]  type:#float 
			checkbox ckbx_scale_aspect_lock "Lock Aspect Ratio" tooltip: "Scales uniformly using U parameters"
		)
		
		group "" (
			spinner spn_channel "Map Channel:" range:[1, 100, 1] type:#integer
			spinner spn_seed "Seed:" range:[0, 99999, 1] type:#integer
		)
		
		fn enableVScale state = (
			spn_v_min_scale.enabled = state
			spn_v_max_scale.enabled = state
		)
		
		on ckbx_scale_aspect_lock changed state do (
			enableVScale (not state)
		) 
		
		on params open do (
			enableVScale (not ckbx_scale_aspect_lock.state)
		)
	) 
	
	rollout attrib "About"
	(
		hyperLink link1 "By Avery Brown" address: "http://averychristianbrown.com/acb/" align:#center color: (color 200 200 200)
		hyperLink link2 "Brooklyn Digital Foundry" address: "https://brooklynfoundry.com/" align:#center color: (color 200 200 200)
		label attrib3 "2019"
	)
	
	parameters data (
		center_offset default:0.5 type:#float animatable:false invisibleInTV:true
	)
	
	-- ** Returns the greater of two inputs **
	fn get_max a b = (if a >= b then return a else return b)
	
	-- ** Compares two floats with a more controllable precision than Maxscript's close_enough **
	fn flt_cmp a b abs_tol = (
		-- Accounts for Max's floating point precision, which is supposedly 1e-7
		local rel_tol = 1e-7
		-- adapted from https://stackoverflow.com/a/33024979
		return ((abs (a-b)) <= get_max ( rel_tol * (get_max (abs a) (abs b) ) ) abs_tol )
	)
	
	-- ** Generates a random float value between a maximum and minimum using an incremental value **
	--  If the maximum and minimum are the same, it will just return the minimum
	-- If the increment is 0.0, it will just return any random value between the bounds
	fn randIncrement minimum maximum increment = (
		if (flt_cmp minimum maximum 1e-4) then return minimum
		
		if (increment < 0.0001) then (
			return (random minimum maximum)
		)
		else (
			-- An integer that contains the exact number of steps between the bounds using the increment
			local x = ceil(abs(maximum - minimum) / increment)
			-- A random integer indicating one of the bounding steps is generated, and then operated on to convert it to the proper float between the bounds
			return ((random 1 x) * increment) + minimum
		)
	)
	
	-- ** Create a translation Matrix in the X and Y (u and v)  axis **
	fn translate_matrix = (
		x = randIncrement u_min_offset u_max_offset offset_inc
		y = randIncrement v_min_offset v_max_offset offset_inc
		return (transMatrix [x, y, 0])
	)
	
	-- ** Create a transform matrix by the center offset forward**
	fn offset_matrix_forward = (
		return (transMatrix [-center_offset, -center_offset, 0])
	)
	
	-- ** Create a transform matrix by the center offset backward**
	fn offset_matrix_backward = (
		return (transMatrix [center_offset,  center_offset, 0])
	)
	
	-- ** Create a rotation Matrix on the z (w) axis **
	fn rotate_matrix = (
		rotationAngle = randIncrement min_rotation max_rotation rotation_inc
		return (rotateZMatrix rotationAngle)
	)

	-- ** Create a scale Matrix in the X and Y (u and v)  axis **
	fn scale_matrix = (
		x = randIncrement u_min_scale u_max_scale scale_inc
		y = randIncrement v_min_scale v_max_scale scale_inc
		return (scaleMatrix  [x, y, 1])
	)
	
	-- ** Create a scale Matrix with the X and Y (u and v) axis locked together**
	fn scale_matrix_locked = (
		x = randIncrement u_min_scale u_max_scale scale_inc
		return (scaleMatrix  [x, x, 1])
	)
	
	-- Adapted from a script by Vojtech Cada (aka Swordslayer) who had, in turn, adapted from another script by Jorge Rodriguez (aka PolyTools3D)
	-- Changed to remove accumulating the position and to output an array of facelists
	fn getMeshElements =
	(
		local elementCount = 0
		local vertCount = mesh.numVerts
		-- Create an array prefilled with all face indices
		local faces = #{1..mesh.numFaces}
		-- Create an array of arrays for all vertices
		local facesByVert = for v = 1 to vertCount collect #()
		local elements_collection = #()
		
		-- For each of the vertices in each face, set the associated face number 
		-- This is stored as an array for each vertex index that contains the index of all associated faces
		-- The result ends up being that accessing a vert index in the facesByVert array returns another array of all faces connected to that vertex
		for face in faces do
		(
			local faceVerts = getFace mesh face
			-- Store each vertex (always 3) for this face
			append facesByVert[faceVerts[1]] face
			append facesByVert[faceVerts[2]] face
			append facesByVert[faceVerts[3]] face
		)
		
		-- For each face index
		for face in faces do
		(
			-- the element containing the face indices (The facelist)
			local element = #(face)
			-- Starts at 0 in order to add one every loop
			local i = 0
			
			-- Loop while there is an unprocessed face in the elements array.  
			-- F is the face index of the current face, which is always the next face in the elements queue.
			
			-- When a face has been processed, it is registered as false in the "faces" array.  This prevents it from being processed again.
			-- This also removes it from the outer loop, causing only the first face of each element to be called there.  Processing a face adds all border faces
			--  to the elements queue ("element" array).  This causes the queued faces to spread across the surface of the element and ends the
			-- loop when all of those faces have been processed.  The faces are continuously added to the queue, sometimes 
			-- more than once.  Duplicate faces are skipped over when they have already been processed once.  When there are no more elements to 
			-- process, the attempt to access the next element will return "undefined," ending the loop.
			
			-- It is faster to return a facelist with duplicates, rather than one with duplicates filtered out.  "meshop.getMapVertsUsingMapFace" 
			-- likely handles the filtering of duplicates much faster than applying "makeUniqueArray " to each facelist.
			
			while (local f = element[i += 1]) != undefined do if faces[f] do
			(
				-- Marks this face as having been processed
				faces[f] = false
				-- gets the vertices from the current face being processed 
				local faceVerts = getFace mesh f
				
				-- First vertex
				-- checks if facesByVert[faceVerts[1]] is an array
				if isKindOf facesByVert[faceVerts[1]] Array do
				(
					-- Adds all faces connected to that vertex to the facelist, duplicates included
					join element facesByVert[faceVerts[1]]
				)
				
				-- Second Vertex
				if isKindOf facesByVert[faceVerts[2]] Array do
				(
					join element facesByVert[faceVerts[2]]
				)
				
				-- Third Vertex
				if isKindOf facesByVert[faceVerts[3]] Array do
				(
					join element facesByVert[faceVerts[3]]
				)
			)
			-- Adds the collected element array (a facelist) to the returning array of facelists
			append elements_collection element
		)
		
		return elements_collection
	)
	
	-- ** Transforms a list of map vertices by a transformation matrix **
	fn transform_vertices map_channel vert_list transform_matrix = (
		for v in vert_list do (
			-- get the mapVert position
			local vtx = meshop.getMapVert mesh map_channel v
			
			-- Multiply by the transform matrix
			vtx *= transform_matrix
			
			-- Set the transformed mapVert
			meshop.setMapVert mesh map_channel v vtx
		)
	)
	
	-- ** In order to remove the evaluation of a bunch of if statements for each element, this function evaluates which operations are 
	--- required and then returns an array of functions that get executed for every element. **
	fn gather_matrix_functions = (
		-- Matrices
		local mats = #()
		-- Center offset forward
		if scale_on or rotation_on do (
			append mats offset_matrix_forward
		)
		
		if scale_on do (
			-- determines if scale should be locked to the X axis
			if scale_aspect_lock then (append mats scale_matrix_locked) else (append mats scale_matrix)
		)
		
		if rotation_on do (
			append mats rotate_matrix
		)
		
		if offset_on do (
			append mats translate_matrix
		)
		
		-- Center offset backward
		if scale_on or rotation_on do (
			append mats offset_matrix_backward
		)
		
		return mats
	)
	
	-- ** Generates and applies the transformation matrix to a list of faces
	fn offset_element index facelist matrix_functions = (
		-- Converts the facelist into mapVerts
		local vert_indices = meshop.getMapVertsUsingMapFace mesh ch facelist
		
		-- The transform for the whole element is created by multiplying several matrices together.
		-- This combined matrix will transform each mapVert to the correct position when multiplied.
		-- The verts are translated so that they rotate around the center of the first uvw tile, rather than the origin at the bottom left corner
		-- The order of operations is inverse of the usual (SRT instead of TRS).  This prevents the scale and rotation from interfering with each other in bizarre ways.
		local t_matrix = transMatrix [0, 0, 0]
		
		-- Multiply generated matrices together from a list of functions
		for m in matrix_functions do (
			mat = m()
			t_matrix *= mat
		)
			
		-- Transform all mapVerts in the vertlist by the combined matrix
		transform_vertices ch vert_indices t_matrix
	)
	
	on modifyMesh do
	(
		if (meshop.getMapSupport mesh ch) then (
			if offset_on or rotation_on or scale_on then (
				
				-- Generate an array of matrix functions as needed
				local matrices = gather_matrix_functions()
				
				seed(rseed)
				-- Gets an array of facelists for each element
				local eles = getMeshElements()
				
				-- Offsets each element with a different random transform Matrix
				for e=1 to eles.count do (
					offset_element e eles[e] matrices
				)
			)
		)
	)
)